# MySQL queries used
Queries = {
    "pl_query": """
        SELECT
            id,
            IFNULL(User, "")                    AS user,
            IFNULL(Host, "")                    AS host,
            IFNULL(db, "")                      AS db,
            IFNULL(Command, "")                 As command,
            IFNULL(Time, "0")                   AS time,
            IFNULL(Info, "")                    AS query,
            IFNULL(State, "")                   AS state,
            IFNULL(trx_query, "")               AS trx_query,
            IFNULL(trx_state, "")               AS trx_state,
            IFNULL(trx_operation_state, "")     AS trx_operation_state,
            IFNULL(trx_rows_locked, "0")        AS trx_rows_locked,
            IFNULL(trx_rows_modified, "0")      AS trx_rows_modified,
            IFNULL(trx_concurrency_tickets, "") AS trx_concurrency_tickets
        FROM
            information_schema.PROCESSLIST pl
            LEFT JOIN information_schema.innodb_trx ON trx_mysql_thread_id = pl.Id
        WHERE 1 $placeholder
    """,
    "ps_query": """
        SELECT
            processlist_id                      AS id,
            IFNULL(processlist_user, "")        AS user,
            IFNULL(processlist_host, "")        AS host,
            IFNULL(processlist_db, "")          AS db,
            IFNULL(processlist_command, "")     As command,
            IFNULL(processlist_time, "0")       AS time,
            IFNULL(processlist_info, "")        AS query,
            IFNULL(processlist_state, "")       AS state,
            IFNULL(trx_query, "")               AS trx_query,
            IFNULL(trx_state, "")               AS trx_state,
            IFNULL(trx_operation_state, "")     AS trx_operation_state,
            IFNULL(trx_rows_locked, "0")        AS trx_rows_locked,
            IFNULL(trx_rows_modified, "0")      AS trx_rows_modified,
            IFNULL(trx_concurrency_tickets, "") AS trx_concurrency_tickets
        FROM
            performance_schema.threads t
            LEFT JOIN information_schema.innodb_trx tx ON trx_mysql_thread_id = t.processlist_id
        WHERE
            processlist_id IS NOT NULL AND
            processlist_time IS NOT NULL AND
            processlist_command != 'Daemon'
            $placeholder
    """,
    "locks_query-5": """
        SELECT
            IFNULL(r.trx_mysql_thread_id, "")                            AS waiting_thread,
            IFNULL(r.trx_query, "")                                      AS waiting_query,
            IFNULL(r.trx_rows_modified, "")                              AS waiting_rows_modified,
            IFNULL(TIMESTAMPDIFF(SECOND, r.trx_started, NOW()), "")      AS waiting_age,
            IFNULL(TIMESTAMPDIFF(SECOND, r.trx_wait_started, NOW()), "") AS waiting_wait_secs,
            IFNULL(b.trx_mysql_thread_id, "")                            AS blocking_thread,
            IFNULL(b.trx_query, "")                                      AS blocking_query,
            IFNULL(b.trx_rows_modified, "")                              AS blocking_rows_modified,
            IFNULL(TIMESTAMPDIFF(SECOND, b.trx_started, NOW()), "")      AS blocking_age,
            IFNULL(TIMESTAMPDIFF(SECOND, b.trx_wait_started, NOW()), "") AS blocking_wait_secs,
            IFNULL(lock_mode, "")                                        AS lock_mode,
            IFNULL(lock_type, "")                                        AS lock_type
        FROM
            INFORMATION_SCHEMA.INNODB_LOCK_WAITS w
            JOIN INFORMATION_SCHEMA.INNODB_TRX b   ON b.trx_id  = w.blocking_trx_id
            JOIN INFORMATION_SCHEMA.INNODB_TRX r   ON r.trx_id  = w.requesting_trx_id
            JOIN INFORMATION_SCHEMA.INNODB_LOCKS l ON l.lock_id = w.requested_lock_id
        ORDER BY
            TIMESTAMPDIFF(SECOND, r.trx_wait_started, NOW()) DESC
    """,
    "locks_query-8": """
        SELECT
            IFNULL(r.trx_mysql_thread_id, "")                            AS waiting_thread,
            IFNULL(r.trx_query, "")                                      AS waiting_query,
            IFNULL(r.trx_rows_modified, "")                              AS waiting_rows_modified,
            IFNULL(TIMESTAMPDIFF(SECOND, r.trx_started, NOW()), "")      AS waiting_age,
            IFNULL(TIMESTAMPDIFF(SECOND, r.trx_wait_started, NOW()), "") AS waiting_wait_secs,
            IFNULL(b.trx_mysql_thread_id, "")                            AS blocking_thread,
            IFNULL(b.trx_query, "")                                      AS blocking_query,
            IFNULL(b.trx_rows_modified, "")                              AS blocking_rows_modified,
            IFNULL(TIMESTAMPDIFF(SECOND, b.trx_started, NOW()), "")      AS blocking_age,
            IFNULL(TIMESTAMPDIFF(SECOND, b.trx_wait_started, NOW()), "") AS blocking_wait_secs,
            IFNULL(lock_mode, "")                                        AS lock_mode,
            IFNULL(lock_type, "")                                        AS lock_type
        FROM
            performance_schema.data_lock_waits w
            JOIN INFORMATION_SCHEMA.INNODB_TRX b ON b.trx_id         = w.blocking_engine_transaction_id
            JOIN INFORMATION_SCHEMA.INNODB_TRX r ON r.trx_id         = w.requesting_engine_transaction_id
            JOIN performance_schema.data_locks l ON l.engine_lock_id = w.requesting_engine_lock_id
        ORDER BY
            TIMESTAMPDIFF(SECOND, r.trx_wait_started, NOW()) DESC
    """,
    "ps_replica_lag": """
        SELECT
            IFNULL(TIMESTAMPDIFF(
                SECOND,
                MIN(APPLYING_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP),
            NOW()), "0") AS secs_behind
        FROM
            performance_schema.replication_applier_status_by_worker
        WHERE
            APPLYING_TRANSACTION != ''
    """,
    "heartbeat_replica_lag": """
        SELECT
            TIMESTAMPDIFF(SECOND, MAX(ts), NOW()) AS secs_behind
        FROM
            $placeholder
    """,
    "ps_find_replicas": """
        SELECT
            processlist_id   AS id,
            processlist_user AS user,
            processlist_host AS host
        FROM
            performance_schema.threads
        WHERE
            processlist_command LIKE 'Binlog Dump%'
    """,
    "pl_find_replicas": """
        SELECT
            Id   AS id,
            User AS user,
            Host AS host
        FROM
            information_schema.PROCESSLIST
        WHERE
            Command Like 'Binlog Dump%'
    """,
    "ps_user_statisitics": """
        SELECT
            u.user AS user,
            total_connections,
            current_connections,
            sum_rows_affected,
            sum_rows_sent,
            sum_rows_examined,
            sum_created_tmp_disk_tables,
            sum_created_tmp_tables
        FROM
            performance_schema.users u
            JOIN performance_schema.events_statements_summary_by_user_by_event_name ess ON u.user = ess.user
        WHERE
            current_connections != 0
        ORDER BY
            current_connections DESC
    """,
    "userstat_user_statisitics": """
        SELECT
            user,
            total_connections,
            concurrent_connections,
            denied_connections,
            binlog_bytes_written,
            rows_fetched,
            rows_updated,
            table_rows_read,
            select_commands,
            update_commands,
            other_commands,
            commit_transactions,
            rollback_transactions,
            access_denied
        FROM
            information_schema.user_statistics
        WHERE
            concurrent_connections != 0
        ORDER BY
            concurrent_connections DESC
    """,
    "error_log": """
        SELECT
            logged AS timestamp,
            prio AS level,
            subsystem,
            data AS message
        FROM
            performance_schema.error_log
        WHERE
            data != 'Could not open log file.'
            $placeholder
        ORDER BY timestamp
    """,
    "memory_by_user": """
        SELECT
            user,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_user_by_current_bytes
        WHERE
            user != "background"
        LIMIT 30
    """,
    "memory_by_code_area": """
        SELECT
            SUBSTRING_INDEX( event_name, '/', 2 ) AS code_area,
            sys.format_bytes (
            SUM( current_alloc )) AS current_allocated
        FROM
            sys.x$memory_global_by_current_bytes
        GROUP BY
            SUBSTRING_INDEX( event_name, '/', 2 )
        ORDER BY
            SUM( current_alloc ) DESC
        LIMIT 30
    """,
    "memory_by_host": """
        SELECT
            host,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_host_by_current_bytes
        WHERE
            host != "background"
        LIMIT 30
    """,
    "databases": """
        SELECT
            SCHEMA_NAME
        FROM
            information_schema.SCHEMATA
        ORDER BY
            SCHEMA_NAME
    """,
    "status": "SHOW GLOBAL STATUS",
    "variables": "SHOW GLOBAL VARIABLES",
    "primary_status": "SHOW MASTER STATUS",
    "replica_status": "SHOW SLAVE STATUS",
    "innodb_status": "SHOW ENGINE INNODB STATUS",
}



import re

from dolphie import Dolphie
from dolphie.Functions import format_number
from rich import box
from rich.style import Style
from rich.table import Table


def create_panel(dolphie: Dolphie) -> Table:
    innodb_lock_threads = {}

    dolphie.db.execute(dolphie.innodb_locks_sql)
    threads = dolphie.db.fetchall()

    for counter, thread in enumerate(threads):
        w_query = re.sub(r"\s+", " ", thread["waiting_query"])
        b_query = re.sub(r"\s+", " ", thread["blocking_query"])

        innodb_lock_threads[counter] = {
            "w_thread": thread["waiting_thread"],
            "w_query": re.sub(r"\s+", " ", w_query),
            "w_rows_modified": thread["waiting_rows_modified"],
            "w_age": thread["waiting_age"],
            "w_wait_secs": thread["waiting_wait_secs"],
            "b_thread": thread["blocking_thread"],
            "b_query": re.sub(r"\s+", " ", b_query),
            "b_rows_modified": thread["blocking_rows_modified"],
            "b_age": thread["blocking_age"],
            "b_wait_secs": thread["blocking_wait_secs"],
            "lock_mode": thread["lock_mode"],
            "lock_type": thread["lock_type"],
        }

    table = Table(header_style="bold white", box=box.SIMPLE_HEAVY, style="steel_blue1")

    columns = {}
    columns["[W](W) Thread ID"] = {"field": "w_thread", "width": 13, "format_number": False}
    columns["[W]Age"] = {"field": "w_age", "width": 4, "format_number": False}
    columns["[W]Wait"] = {"field": "w_wait_secs", "width": 5, "format_number": False}
    columns["[W]Query"] = {"field": "w_query", "width": None, "format_number": False}
    columns["[B](B) Thread ID"] = {"field": "b_thread", "width": 13, "format_number": False}
    columns["[B]Age"] = {"field": "b_age", "width": 4, "format_number": False}
    columns["[B]Wait"] = {"field": "b_wait_secs", "width": 5, "format_number": False}
    columns["[B]Rows Mod"] = {"field": "b_rows_modified", "width": 8, "format_number": True}
    columns["[B]Query"] = {"field": "b_query", "width": None, "format_number": False}
    columns["Mode"] = {"field": "lock_mode", "width": 7, "format_number": False}
    columns["Type"] = {"field": "lock_type", "width": 8, "format_number": False}

    total_width = 0
    for column, data in columns.items():
        if "Query" in column:
            overflow = "crop"
        else:
            overflow = "ellipsis"

        if "[B]" in column:
            column = column.replace("[B]", "")
            row_style = Style(color="red")
        elif "[W]" in column:
            column = column.replace("[W]", "")
            row_style = Style(color="magenta")
        else:
            row_style = Style(color="grey93")

        if data["width"]:
            table.add_column(column, width=data["width"], no_wrap=True, header_style=row_style, overflow=overflow)
            total_width += data["width"]
        else:
            table.add_column(column, no_wrap=True, header_style=row_style, overflow=overflow)

    # This variable is to cut off query so it's the perfect width for the auto-sized column that matches terminal width
    query_characters = round((dolphie.console.size.width / 2) - ((len(columns) * 4) + 6))

    for id, thread in innodb_lock_threads.items():
        b_wait_secs = "0s"
        if thread["b_wait_secs"]:
            b_wait_secs = "%ss" % thread["b_wait_secs"]

        w_wait_secs = "0s"
        if thread["w_wait_secs"]:
            w_wait_secs = "%ss" % thread["w_wait_secs"]

        waiting_query = thread["w_query"]
        if not waiting_query:
            waiting_query = waiting_query.ljust(query_characters)

        elif len(waiting_query) < query_characters:
            waiting_query = waiting_query.ljust(dolphie.console.size.width)

        blocking_query = thread["b_query"]
        if not blocking_query:
            blocking_query = blocking_query.ljust(query_characters)

        elif len(blocking_query) < query_characters:
            blocking_query = blocking_query.ljust(dolphie.console.size.width)

        # Change values to what we want
        thread["w_age"] = "%ss" % thread["w_age"]
        thread["b_age"] = "%ss" % thread["b_age"]
        thread["w_query"] = waiting_query[0:query_characters]
        thread["b_query"] = blocking_query[0:query_characters]
        thread["w_wait_secs"] = w_wait_secs
        thread["b_wait_secs"] = b_wait_secs

        # Add rows to the table
        row_values = []
        for column, data in columns.items():
            if data["format_number"]:
                row_values.append(format_number(thread[data["field"]]))
            else:
                row_values.append(thread[data["field"]])

        table.add_row(*row_values, style="grey93")

    # Add an invisible row to keep query columns sized correctly
    if len(innodb_lock_threads) == 0:
        empty_values = []
        for column, data in columns.items():
            if data["width"]:
                empty_values.append("")
            else:
                empty_values.append("".ljust(query_characters))

        table.add_row(*empty_values)

    return table



#!/usr/bin/env python3

# ****************************
# *        Dolphie           *
# * Author: Charles Thompson *
# ****************************

import os
import re
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser
from datetime import datetime
from time import sleep

import myloginpath
from dolphie import Dolphie
from dolphie.ManualException import ManualException
from dolphie.Panels import (
    dashboard_panel,
    innodb_io_panel,
    innodb_locks_panel,
    processlist_panel,
    replica_panel,
)
from dolphie.Queries import Queries
from rich.live import Live
from rich.prompt import Prompt


def parse_args(dolphie: Dolphie):
    epilog = """
Config file with [client] section supports these options:
    host
    user
    password
    port
    socket
    ssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
    ssl_ca
    ssl_cert
    ssl_key

Login path file supports these options:
    host
    user
    password
    port
    socket

Environment variables support these options:
    DOLPHIE_USER
    DOLPHIE_PASSWORD
    DOLPHIE_HOST
    DOLPHIE_PORT
    DOLPHIE_SOCKET

"""
    parser = ArgumentParser(
        conflict_handler="resolve",
        description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
        epilog=epilog,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-u",
        "--user",
        dest="user",
        type=str,
        help="Username for MySQL",
    )
    parser.add_argument("-p", "--password", dest="password", type=str, help="Password for MySQL")
    parser.add_argument(
        "--ask-pass",
        dest="ask_password",
        action="store_true",
        default=False,
        help="Ask for password (hidden text)",
    )
    parser.add_argument(
        "-h",
        "--host",
        dest="host",
        type=str,
        help="Hostname/IP address for MySQL",
    )
    parser.add_argument(
        "-P",
        "--port",
        dest="port",
        type=int,
        help="Port for MySQL (Socket has precendence)",
    )
    parser.add_argument(
        "-S",
        "--socket",
        dest="socket",
        type=str,
        help="Socket file for MySQL",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=str,
        help=(
            "Absolute config file path to use. This should use [client] section. "
            "See below for options support [default: ~/.my.cnf]"
        ),
    )
    parser.add_argument(
        "-f",
        "--host-cache-file",
        dest="host_cache_file",
        type=str,
        help=(
            "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
            "using format: ip=hostname [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-l",
        "--login-path",
        dest="login_path",
        default="client",
        type=str,
        help=(
            "Specify login path to use mysql_config_editor's file ~/.mylogin.cnf for encrypted login credentials. "
            "Supercedes config file [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-r",
        "--refresh_interval",
        dest="refresh_interval",
        default=1,
        type=int,
        help="How much time to wait in seconds between each refresh [default: %(default)s]",
    )
    parser.add_argument(
        "-R",
        "--refresh_interval_innodb_status",
        dest="refresh_interval_innodb_status",
        default=1,
        type=int,
        help=(
            "How much time to wait in seconds to execute SHOW ENGINE INNODB STATUS to refresh data its responsible "
            "for [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-H",
        "--heartbeat-table",
        dest="heartbeat_table",
        type=str,
        help=(
            "If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
            "has for replication lag instead of Seconds_Behind_Master from SHOW SLAVE STATUS"
        ),
    )
    parser.add_argument(
        "--ssl-mode",
        dest="ssl_mode",
        type=str,
        help=(
            "Desired security state of the connection to the host. Supports: "
            "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
        ),
    )
    parser.add_argument(
        "--ssl-ca",
        dest="ssl_ca",
        type=str,
        help="Path to the file that contains a PEM-formatted CA certificate",
    )
    parser.add_argument(
        "--ssl-cert",
        dest="ssl_cert",
        type=str,
        help="Path to the file that contains a PEM-formatted client certificate",
    )
    parser.add_argument(
        "--ssl-key",
        dest="ssl_key",
        type=str,
        help="Path to the file that contains a PEM-formatted private key for the client certificate",
    )
    parser.add_argument(
        "--hide-dashboard",
        dest="dashboard",
        action="store_false",
        default=True,
        help=(
            "Start without showing dashboard. This is good to use if you want to reclaim terminal space and "
            "not execute the additional queries for it"
        ),
    )
    parser.add_argument(
        "--show-trxs-only",
        dest="show_trxs_only",
        action="store_true",
        default=False,
        help="Start with only showing queries that are running a transaction",
    )
    parser.add_argument(
        "--additional-columns",
        dest="show_additional_query_columns",
        action="store_true",
        default=False,
        help="Start with additional columns in processlist panel",
    )
    parser.add_argument(
        "--use-processlist",
        dest="use_processlist",
        action="store_true",
        default=False,
        help="Start with using Processlist instead of Performance Schema for listing queries",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=dolphie.app_version, help="Display version and exit"
    )

    parameter_options = vars(parser.parse_args())  # Convert object to dict
    basic_options = ["user", "password", "host", "port", "socket"]

    # Use specified config file if there is one, else use standard ~/.my.cnf
    if parameter_options["config_file"]:
        dolphie.config_file = parameter_options["config_file"]
    else:
        dolphie.config_file = "%s/.my.cnf" % os.path.expanduser("~")

    # Use config file for login credentials
    if os.path.isfile(dolphie.config_file):
        cfg = ConfigParser()
        cfg.read(dolphie.config_file)

        for option in basic_options:
            if cfg.has_option("client", option):
                setattr(dolphie, option, cfg.get("client", option))

        if cfg.has_option("client", "ssl_mode"):
            ssl_mode = cfg.get("client", "ssl_mode").upper()

            if ssl_mode == "REQUIRED":
                dolphie.ssl[""] = True
            elif ssl_mode == "VERIFY_CA":
                dolphie.ssl["check_hostname"] = False
            elif ssl_mode == "VERIFY_IDENTITY":
                dolphie.ssl["check_hostname"] = True
            else:
                raise ManualException("Unsupported SSL mode [b]%s" % ssl_mode)

        if cfg.has_option("client", "ssl_ca"):
            dolphie.ssl["ca"] = cfg.get("client", "ssl_ca")
        if cfg.has_option("client", "ssl_cert"):
            dolphie.ssl["cert"] = cfg.get("client", "ssl_cert")
        if cfg.has_option("client", "ssl_key"):
            dolphie.ssl["key"] = cfg.get("client", "ssl_key")

    # Use login path for login credentials
    if parameter_options["login_path"]:
        try:
            login_path_data = myloginpath.parse(parameter_options["login_path"])

            for option in basic_options:
                if option in login_path_data:
                    setattr(dolphie, option, login_path_data[option])
        except Exception as e:
            # Don't error out for default login path
            if parameter_options["login_path"] != "client":
                raise ManualException("Problem reading login path file", reason=e)

    # Use environment variables for basic options if specified
    for option in basic_options:
        environment_var = "DOLPHIE_%s" % option.upper()
        if environment_var in os.environ and os.environ[environment_var]:
            setattr(dolphie, option, os.environ[environment_var])

    # Lastly, use parameter options if specified
    for option in basic_options:
        if parameter_options[option]:
            setattr(dolphie, option, parameter_options[option])

    if parameter_options["ask_password"]:
        dolphie.password = Prompt.ask("[b steel_blue1]Password", password=True)

    if not dolphie.host:
        dolphie.host = "localhost"

    if parameter_options["refresh_interval"]:
        dolphie.refresh_interval = parameter_options["refresh_interval"]

    if parameter_options["refresh_interval_innodb_status"]:
        dolphie.refresh_interval_innodb_status = parameter_options["refresh_interval_innodb_status"]

    if parameter_options["heartbeat_table"]:
        pattern_match = re.search(r"^(\w+\.\w+)$", parameter_options["heartbeat_table"])
        if pattern_match:
            dolphie.heartbeat_table = parameter_options["heartbeat_table"]
            Queries["heartbeat_replica_lag"] = Queries["heartbeat_replica_lag"].replace(
                "$placeholder", dolphie.heartbeat_table
            )
        else:
            raise ManualException("Your heartbeat table did not conform to the proper format: db.table")

    if parameter_options["ssl_mode"]:
        ssl_mode = parameter_options["ssl_mode"].upper()

        if ssl_mode == "REQUIRED":
            dolphie.ssl[""] = True
        elif ssl_mode == "VERIFY_CA":
            dolphie.ssl["check_hostame"] = False
        elif ssl_mode == "VERIFY_IDENTITY":
            dolphie.ssl["check_hostame"] = True
        else:
            raise ManualException(f"Unsupported SSL mode {ssl_mode}")

    if parameter_options["ssl_ca"]:
        dolphie.ssl["ca"] = parameter_options["ssl_ca"]
    if parameter_options["ssl_cert"]:
        dolphie.ssl["cert"] = parameter_options["ssl_cert"]
    if parameter_options["ssl_key"]:
        dolphie.ssl["key"] = parameter_options["ssl_key"]

    if parameter_options["host_cache_file"]:
        dolphie.host_cache_file = parameter_options["host_cache_file"]
    else:
        dolphie.host_cache_file = os.path.dirname(os.path.abspath(__file__)) + "/host_cache"

    dolphie.dashboard = parameter_options["dashboard"]
    dolphie.show_trxs_only = parameter_options["show_trxs_only"]
    dolphie.show_additional_query_columns = parameter_options["show_additional_query_columns"]
    dolphie.use_processlist = parameter_options["use_processlist"]

    if dolphie.dashboard:
        dolphie.layout["dashboard"].visible = True
    else:
        dolphie.layout["dashboard"].visible = False


def main():
    dolphie = Dolphie()

    try:
        dolphie.create_rich_layout()
        parse_args(dolphie)
        dolphie.check_for_update()
        dolphie.db_connect()
        dolphie.load_host_cache_file()
        dolphie.update_environment_variables_for_pager()

        with Live(
            dolphie.layout, vertical_overflow="crop", screen=True, transient=True, auto_refresh=False
        ) as dolphie.rich_live:
            while True:
                if dolphie.pause_refresh is False:
                    loop_time = datetime.now()

                    dolphie.statuses = dolphie.fetch_data("status")
                    if dolphie.first_loop:
                        dolphie.saved_status = dolphie.statuses.copy()

                    dolphie.loop_duration_seconds = (loop_time - dolphie.previous_main_loop_time).total_seconds()
                    loop_duration_innodb_status_seconds = (
                        loop_time - dolphie.previous_innodb_status_loop_time
                    ).total_seconds()

                    if dolphie.layout["processlist"].visible:
                        dolphie.processlist_threads = processlist_panel.get_data(dolphie)
                        dolphie.layout["processlist"].update(processlist_panel.create_panel(dolphie))

                    if dolphie.dashboard:
                        dolphie.variables = dolphie.fetch_data("variables")
                        dolphie.primary_status = dolphie.fetch_data("primary_status")
                        dolphie.replica_status = dolphie.fetch_data("replica_status")

                        if (
                            dolphie.first_loop
                            or loop_duration_innodb_status_seconds >= dolphie.refresh_interval_innodb_status
                        ):
                            dolphie.innodb_status = dolphie.fetch_data("innodb_status")

                        dolphie.layout["dashboard"].update(dashboard_panel.create_panel(dolphie))

                        # Save some variables to be used in next refresh
                        dolphie.previous_binlog_position = 0
                        if dolphie.primary_status:
                            dolphie.previous_binlog_position = dolphie.primary_status["Position"]

                    if dolphie.layout["replicas"].visible:
                        dolphie.layout["replicas"].update(replica_panel.create_panel(dolphie))

                    if dolphie.layout["innodb_io"].visible:
                        dolphie.layout["innodb_io"].update(innodb_io_panel.create_panel(dolphie))

                    if dolphie.layout["innodb_locks"].visible:
                        dolphie.layout["innodb_locks"].update(innodb_locks_panel.create_panel(dolphie))

                    if dolphie.replica_status:
                        dolphie.previous_replica_sbm = 0
                        if dolphie.replica_status["Seconds_Behind_Master"] is not None:
                            dolphie.previous_replica_sbm = dolphie.replica_status["Seconds_Behind_Master"]

                    # This is for the many stats per second in Dolphie
                    dolphie.saved_status = dolphie.statuses.copy()
                    dolphie.previous_main_loop_time = loop_time

                    if loop_duration_innodb_status_seconds >= dolphie.refresh_interval_innodb_status:
                        dolphie.previous_innodb_status_loop_time = loop_time

                    dolphie.rich_live.update(dolphie.layout, refresh=True)
                else:
                    # To prevent main loop from eating up 100% CPU
                    sleep(0.01)

                # Detect a keypress loop
                loop_counter = 0
                while loop_counter <= dolphie.refresh_interval * 10:
                    if dolphie.kb.key_press():
                        if dolphie.pause_refresh is False:
                            key = dolphie.kb.getch()

                        if key:
                            valid_key = dolphie.capture_key(key)
                            if valid_key:
                                break

                    # refresh_interval * 10 * 100ms equals to the second of sleep a user wants. This allows us to
                    # capture a key faster than sleeping the whole second of refresh_interval
                    sleep(0.100)

                    loop_counter += 1

                dolphie.first_loop = False
    except ManualException as e:
        dolphie.console.print(e.output())
    except Exception:
        dolphie.console.print_exception()


if __name__ == "__main__":
    main()



import re
import string
from datetime import timedelta

from dolphie import Dolphie
from dolphie.Functions import format_number
from dolphie.Queries import Queries
from rich import box
from rich.table import Table


def create_panel(dolphie: Dolphie) -> Table:
    table = Table(header_style="bold white", box=box.SIMPLE_HEAVY, style="steel_blue1")

    columns = {}
    columns["Thread ID"] = {"field": "id", "width": 11, "format_number": False}
    columns["Username"] = {"field": "user", "width": 13, "format_number": False}

    if dolphie.show_additional_query_columns:
        columns["Hostname/IP"] = {"field": "host", "width": 16, "format_number": False}
        columns["Database"] = {"field": "db", "width": 13, "format_number": False}

    columns["Command"] = {"field": "command", "width": 8, "format_number": False}
    columns["State"] = {"field": "state", "width": 16, "format_number": False}
    columns["TRX State"] = {"field": "trx_state", "width": 9, "format_number": False}
    columns["Rows Lock"] = {"field": "trx_rows_locked", "width": 9, "format_number": True}
    columns["Rows Mod"] = {"field": "trx_rows_modified", "width": 8, "format_number": True}

    if (
        dolphie.show_additional_query_columns
        and "innodb_thread_concurrency" in dolphie.variables
        and dolphie.variables["innodb_thread_concurrency"]
    ):
        columns["Tickets"] = {"field": "trx_concurrency_tickets", "width": 8, "format_number": True}

    columns["Time"] = {"field": "formatted_time", "width": 9, "format_number": False}
    columns["Query"] = {"field": "shortened_query", "width": None, "format_number": False}

    total_width = 0
    for column, data in columns.items():
        if column == "Query":
            overflow = "crop"
        else:
            overflow = "ellipsis"

        if data["width"]:
            table.add_column(column, width=data["width"], no_wrap=True, overflow=overflow)
            total_width += data["width"]
        else:
            table.add_column(column, no_wrap=True, overflow=overflow)

    # This variable is to cut off query so it's the perfect width for the auto-sized column that matches terminal width
    query_characters = dolphie.console.size.width - total_width - ((len(columns) * 3) + 1)

    thread_counter = 0
    for id, thread in dolphie.processlist_threads.items():
        query = re.sub(r"\s+", " ", thread["query"])

        # Replace strings with [NONPRINTABLE] if they contain non-printable characters
        for m in re.findall(r"(\"(?:(?!(?<!\\)\").)*\"|'(?:(?!(?<!\\)').)*')", query):
            test_pattern = re.search(f"[^{re.escape(string.printable)}]", m)
            if test_pattern:
                query = query.replace(m, "[NONPRINTABLE]")

        # If no query, pad the query column with spaces so it's sized correctly
        if not query:
            query = query.ljust(query_characters)

        # Pad queries with spaces that are not the full size of query column
        elif len(query) < query_characters:
            query = query.ljust(dolphie.console.size.width)

        # Change values to what we want
        if thread["command"] == "Killed":
            thread["command"] = "[bright_red]%s" % thread["command"]
        else:
            thread["command"] = thread["command"]

        thread["shortened_query"] = query[0:query_characters]

        # Add rows for each thread
        row_values = []
        for column, data in columns.items():
            if data["format_number"]:
                row_values.append(format_number(thread[data["field"]]))
            else:
                row_values.append(thread[data["field"]])

        table.add_row(*row_values, style="grey93")

        thread_counter += 1

    # Add an invisible row to keep query column sized correctly
    if thread_counter == 0:
        empty_values = []
        for column, data in columns.items():
            if data["width"]:
                empty_values.append("")
            else:
                empty_values.append("".ljust(query_characters))

        table.add_row(*empty_values)

    return table


def get_data(dolphie: Dolphie):
    if dolphie.use_performance_schema:
        processlist_query = Queries["ps_query"]
    else:
        processlist_query = Queries["pl_query"]

    if dolphie.sort_by_time_descending:
        if dolphie.use_performance_schema:
            processlist_query = processlist_query + " ORDER BY processlist_time DESC"
        else:
            processlist_query = processlist_query + " ORDER BY LENGTH(Time) DESC, Time DESC"
    else:
        if dolphie.use_performance_schema:
            processlist_query = processlist_query + " ORDER BY processlist_time"
        else:
            processlist_query = processlist_query + " ORDER BY LENGTH(Time), Time"

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Only show running queries
    if not dolphie.show_idle_queries:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_command != 'Sleep' AND processlist_command NOT LIKE 'Binlog Dump%') AND "
                "(processlist_info IS NOT NULL OR trx_query IS NOT NULL)"
            )
        else:
            where_clause.append(
                "(Command != 'Sleep' AND Command NOT LIKE 'Binlog Dump%') AND "
                "(Info IS NOT NULL OR trx_query IS NOT NULL)"
            )

    # Only show running transactions only
    if dolphie.show_trxs_only:
        where_clause.append("trx_state != ''")

    # Filter user
    if dolphie.user_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_user = '%s'" % dolphie.user_filter)
        else:
            where_clause.append("User = '%s'" % dolphie.user_filter)

    # Filter database
    if dolphie.db_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_db = '%s'" % dolphie.db_filter)
        else:
            where_clause.append("db = '%s'" % dolphie.db_filter)

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        if dolphie.use_performance_schema:
            where_clause.append("processlist_host LIKE '%s%%'" % dolphie.host_filter)
        else:
            where_clause.append("Host LIKE '%s%%'" % dolphie.host_filter)

    # Filter time
    if dolphie.time_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_time >= '%s'" % dolphie.time_filter)
        else:
            where_clause.append("Time >= '%s'" % dolphie.time_filter)

    # Filter query
    if dolphie.query_filter:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_info LIKE '%%%s%%' OR trx_query LIKE '%%%s%%')"
                % (dolphie.query_filter, dolphie.query_filter),
            )
        else:
            where_clause.append("Info LIKE '%%%s%%'" % dolphie.query_filter)

    # Add in our dynamic WHERE clause for filtering
    if where_clause:
        processlist_query = processlist_query.replace("$placeholder", "AND " + " AND ".join(where_clause))
    else:
        processlist_query = processlist_query.replace("$placeholder", "")

    # Limit the SELECT query to only retrieve how many lines the terminal has
    processlist_query = processlist_query + " LIMIT %s" % (dolphie.console.size.height)

    processlist_threads = {}
    # Run the processlist query
    dolphie.db.execute(processlist_query)
    threads = dolphie.db.fetchall()

    for thread in threads:
        # Don't include Dolphie's thread
        if dolphie.connection_id == thread["id"]:
            continue

        command = thread["command"]
        if dolphie.use_performance_schema and dolphie.show_last_executed_query is False and command == "Sleep":
            query = ""
        else:
            # Use trx_query over Performance Schema query since it's more accurate
            if dolphie.use_performance_schema and thread["trx_query"]:
                query = thread["trx_query"]
            else:
                query = thread["query"]

        # Determine time color
        time = int(thread["time"])
        thread_color = "grey93"
        if "SELECT /*!40001 SQL_NO_CACHE */ *" in query:
            thread_color = "magenta"
        elif query and command != "Sleep" and "Binlog Dump" not in command:
            if time >= 4:
                thread_color = "bright_red"
            elif time >= 2:
                thread_color = "bright_yellow"
            elif time <= 1:
                thread_color = "bright_green"

        formatted_time = "[%s]%ss" % (thread_color, time)

        # If after the first loop there's nothing in cache, don't try to resolve anymore.
        # This is an optimization
        host = thread["host"].split(":")[0]
        if dolphie.first_loop is False:
            if dolphie.host_cache:
                host = dolphie.get_hostname(host)
        else:
            host = dolphie.get_hostname(host)

        processlist_threads[str(thread["id"])] = {
            "id": str(thread["id"]),
            "user": thread["user"],
            "host": host,
            "db": thread["db"],
            "formatted_time": formatted_time,
            "time": time,
            "hhmmss_time": "[{}]{:0>8}".format(thread_color, str(timedelta(seconds=time))),
            "command": command,
            "state": thread["state"],
            "trx_state": thread["trx_state"],
            "trx_operation_state": thread["trx_operation_state"],
            "trx_rows_locked": thread["trx_rows_locked"],
            "trx_rows_modified": thread["trx_rows_modified"],
            "trx_concurrency_tickets": thread["trx_concurrency_tickets"],
            "query": query,
        }

    return processlist_threads



from rich import box
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table


class ManualException(Exception):
    def __init__(self, message, query="", reason=""):
        self.message = message
        self.reason = reason
        self.query = query

    def output(self):
        table_exception = Table(
            box=box.SIMPLE_HEAVY,
            style="grey70",
            title="  Dolphie :dolphin:",
            title_style=Style(color="grey93", bold=True),
            title_justify="left",
            header_style=Style(color="indian_red", bold=True),
            show_header=True,
        )

        table_exception.add_column("Error", justify="left")
        table_exception.add_row(self.message)

        if self.query:
            self.query = Syntax(
                self.query,
                "sql",
                line_numbers=False,
                word_wrap=True,
                theme="monokai",
                background_color="default",
            )
            table_exception.add_row(self.query)

        table_exception.add_row("")

        if self.reason:
            table_exception.add_row("[indian_red]%s" % self.reason)

        return table_exception



import pymysql
from dolphie.Functions import detect_encoding
from dolphie.ManualException import ManualException


class Database:
    def __init__(self, host, user, password, socket, port, ssl):
        self.host = host
        self.user = user
        self.password = password
        self.socket = socket
        self.port = port
        self.ssl = ssl

        try:
            self.connection = pymysql.connect(
                host=host,
                user=user,
                passwd=password,
                unix_socket=socket,
                port=port,
                use_unicode=False,
                ssl=ssl,
                autocommit=True,
            )
            self.cursor = self.connection.cursor(pymysql.cursors.DictCursor)
        except pymysql.Error as e:
            raise ManualException("Failed to connect to database host %s" % self.host, reason=e.args[1])
        except FileNotFoundError:  # Catch SSL file path errors
            raise ManualException("SSL certificate file path isn't valid!")

    def execute(self, query, values=None, ignore_error=False):
        # Prefix all queries with dolphie so they can be identified in the processlist
        query = "/* dolphie */ " + query

        try:
            return self.cursor.execute(query, values)
        except Exception as e:
            if ignore_error:
                return None
            else:
                raise ManualException("Failed to execute query\n", query=query, reason=e.args[1])

    def fetchall(self):
        rows = []

        # Iterate over each row obtained from self.cursor.fetchall()
        for row in self.cursor.fetchall():
            processed_row = {}

            # Iterate over each field-value pair in the row dictionary
            for field, value in row.items():
                if isinstance(value, (bytes, bytearray)):
                    # If the value is an instance of bytes or bytearray, decode it
                    if "query" in field:
                        # If the field name contains the word 'query', detect the encoding
                        processed_row[field] = value.decode(detect_encoding(value))
                    else:
                        # Otherwise, decode the value as utf-8 by default
                        processed_row[field] = value.decode()
                else:
                    # Otherwise, use the original value
                    processed_row[field] = value

            # Append the processed row to the rows list
            rows.append(processed_row)

        # Return the list of processed rows
        return rows

    def fetchone(self, query, field, values=None):
        self.execute(query, values)
        data = self.cursor.fetchone()

        if not data:
            return None

        value = data[field]
        if isinstance(value, (bytes, bytearray)):
            # If the value is an instance of bytes or bytearray, decode it
            if field == "Status":
                # If the field name is Status, detect the encoding
                return value.decode(detect_encoding(value))
            else:
                # Otherwise, decode the value as utf-8 by default
                return value.decode()
        else:
            return value

    def close(self):
        if self.connection:
            self.connection.close()
